"""Upload a wasm file to a Free-Wili."""

import pathlib

from freewili import FreeWili

# The fwi file to upload
my_fwi_file = pathlib.Path(r"~/path/to/MyFile.fwi")
# The wasm file to upload
my_wasm_file = pathlib.Path(r"~/path/to/MyFile.wasm")

# Find connected Free-Wilis
devices = FreeWili.find_all()
if not devices:
    print("No Free-Wili devices found!")
    exit(1)

# Pick the first Free-Wili
device = devices[0]

# We can leave target_name and processor None and it will automatically
# figure out where to send the file.
device.send_file(my_fwi_file, None, None).expect("Failed to upload file")
device.send_file(my_wasm_file, None, None).expect("Failed to upload file")
device.run_script(my_wasm_file.name).expect("Failed to run script")



"""Read buttons on a Free-Wili."""

from result import Err, Ok

from freewili import FreeWili
from freewili.framing import ResponseFrame


def check_resp(device: FreeWili, resp: ResponseFrame) -> None:
    """Check the response and exit on error."""
    if not resp.is_ok():
        print(f"Response failed: {resp.response}")
        device.close()
        exit(1)


BUTTON_COLOR = ["White", "Yellow", "Green", "Blue", "Red"]

# find a FreeWili device
match FreeWili.find_first():
    case Ok(d):
        device = d
        device.stay_open = True
        print(f"Using {device}")
    case Err(msg):
        print(msg)
        exit(1)

try:
    # Read the buttons and print on change
    print("Reading buttons...")
    last_read = bytes([0, 0, 0, 0, 0])
    while True:
        try:
            resp = device.read_all_buttons().expect("Failed to read buttons")
            check_resp(device, resp)
            buttons = resp.response_as_bytes().expect("Failed to convert response")
            for i, button in enumerate(buttons):
                if last_read[i] != button:
                    state = "Pressed \N{WHITE HEAVY CHECK MARK}"
                    if button == 0:
                        state = "Released \N{CROSS MARK}"
                    print(f"{BUTTON_COLOR[i]} {state}")
            last_read = buttons
        except KeyboardInterrupt:
            break
finally:
    device.close()



"""Support for FreeWili serial framing."""

import enum
from dataclasses import dataclass
from typing import Self

import numpy as np
from result import Err, Ok, Result


class ResponseFrameType(enum.Enum):
    """FreeWili serial response frame types."""

    # response frames that start with a menu letter. i\w = i2c write
    Standard = enum.auto()
    # response frames that start with an *[EVENT_NAME]
    Event = enum.auto()


@dataclass(frozen=True)
class ResponseFrame:
    """Response Frame from a FreeWili serial."""

    rf_type: ResponseFrameType
    rf_type_data: str
    # unix epoch timestamp in nanoseconds
    timestamp: int
    seq_number: int
    response: str
    success: int
    _raw: str

    @classmethod
    def from_raw(cls, frame: str, strict: bool = True) -> Result[Self, str]:
        """Take a response frame string and create a ResponseFrame.

        Parameters:
        ----------
            frame : str
                response frame string to decode.
            strict : bool
                allows the timestamp to be invalid.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if decoded successfully, Err(str) if not.
        """
        # Verify we are a frame, we should be enclosed with []
        raw = frame
        if not frame.startswith("[") and not frame.endswith("]"):
            return Err("Invalid data, expected frame to be enclosed []")
        # Strip the brackets
        frame = frame.lstrip("[").rstrip().rstrip("]")
        # Seperate the frame, every item is space seperated.
        items = frame.split(" ")
        if len(items) < 5:
            return Err("Invalid frame contents, not enough items")
        # example I2C: [i\w YXNkZmFzZGY= 4 Invalid 0]
        rf_type = ResponseFrameType.Standard
        if items[0].startswith("*"):
            rf_type = ResponseFrameType.Event
            items[0] = items[0].lstrip("*")
        try:
            # Convert the timestamp to nanoseconds.
            # value is sent as big endian hex.
            big_endian_hex_ts = items[1]
            # unix_epoch_nanoseconds = int(little_endian_hex_ts, 16)
            unix_epoch_nanoseconds = int(big_endian_hex_ts, 16)
        except ValueError as ex:
            if strict:
                return Err(f"Failed to decode timestamp: {str(ex)}")
            unix_epoch_nanoseconds = 0
        return Ok(
            cls(
                rf_type,  # rf_type
                str(items[0]),  # rf_type_data
                unix_epoch_nanoseconds,  # timestamp
                int(items[2]),  # seq_number
                " ".join(items[3:-1]),  # response
                int(items[-1]),  # success
                raw,  # _raw
            )
        )

    def is_ok(self) -> bool:
        """Validates if the frame was successful.

        Parameters:
        ----------
            None

        Returns:
        -------
            bool:
                True if success == 1, False otherwise.
        """
        return self.success == 1

    def response_as_bytes(self, check_ok: bool = True) -> Result[bytes, str]:
        """Convert the response into bytes.

        Parameters:
        ----------
            check_ok: bool
                Calls is_ok() to make sure the frame is valid.

        Returns:
        -------
            Result[bytes, str]:
                Ok(bytes) if valid, False if data couldn't be converted.
        """
        if check_ok and not self.is_ok():
            return Err("Response success is not ok")
        try:
            response = self.response.rstrip(" ")
            if not response:
                return Ok(bytes())
            data: bytes = bytes([int(x, 16) for x in response.split(" ")])
            return Ok(data)
        except TypeError as ex:
            return Err(ex)
        except ValueError as ex:
            return Err(ex)

    def timestamp_as_datetime(self, check_ok: bool = False) -> Result[np.datetime64, str]:
        """Convert the timestamp into a datetime.

        Parameters:
        ----------
            check_ok: bool
                Calls is_ok() to make sure the frame is valid.

        Returns:
        -------
            Result[np.datetime64, str]:
                Ok(np.datetime64) if valid, Err(str) if timestamp couldn't be converted.
        """
        if check_ok and not self.is_ok():
            return Err("Response success is not ok")
        return Ok(np.datetime64(self.timestamp, "ns"))



"""Common data types and constants."""

import enum


class FreeWiliProcessorType(enum.Enum):
    """Processor type of the Free-Wili."""

    Main = enum.auto()
    MainUF2 = enum.auto()
    Display = enum.auto()
    DisplayUF2 = enum.auto()
    FTDI = enum.auto()
    Unknown = enum.auto()

    def __str__(self) -> str:
        return self.name



"""Command line interface for converting images.

This module provides a command line interface to find and control FreeWili boards.
"""

import argparse
import importlib.metadata

from result import Err, Ok

from freewili import image
from freewili.cli import exit_with_error


def main() -> None:
    """A command line interface to convert a jpg or png image to a fwi file.

    Parameters:
    ----------
        None

    Returns:
    -------
        None
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-i",
        "--input",
        required=True,
        help="Path to a JPG or PNG image to be converted",
    )
    parser.add_argument(
        "-o",
        "--output",
        required=True,
        help="Output filename for the fwi file",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=importlib.metadata.version("freewili")),
    )
    args = parser.parse_args()
    match image.convert(args.input, args.output):
        case Ok(msg):
            print(msg)
        case Err(msg):
            exit_with_error(msg)


if __name__ == "__main__":
    main()



"""Toggle IO on a Free-Wili."""

from result import Err, Ok

from freewili import FreeWili
from freewili.framing import ResponseFrame
from freewili.serial_util import IOMenuCommand


def check_resp(device: FreeWili, resp: ResponseFrame) -> None:
    """Check the response and exit on error."""
    if not resp.is_ok():
        print(f"Response failed: {resp.response}")
        device.close()
        exit(1)


# find a FreeWili device
match FreeWili.find_first():
    case Ok(d):
        device = d
        device.stay_open = True
        print(f"Using {device}")
    case Err(msg):
        print(msg)
        exit(1)

try:
    # Set IO 25 high
    resp = device.set_io(25, IOMenuCommand.High).expect("Failed to set IO high")
    check_resp(device, resp)

    # Set IO 25 Low
    resp = device.set_io(25, IOMenuCommand.Low).expect("Failed to set IO low")
    check_resp(device, resp)

    # Toggle IO 25 Low
    resp = device.set_io(25, IOMenuCommand.Toggle).expect("Failed to toggle IO")
    check_resp(device, resp)

    # PWM IO 25
    resp = device.set_io(25, IOMenuCommand.Pwm, 10, 50).expect("Failed to toggle IO")
    check_resp(device, resp)
finally:
    device.close()



"""Test code for freewili.fw module."""

import pytest

from freewili.fw import FileMap
from freewili.serial_util import FreeWiliProcessorType


def test_file_mappings() -> None:
    """Test file mapping."""
    known_maps = {
        "wasm": (FreeWiliProcessorType.Main, "/scripts", "WASM binary"),
        "wsm": (FreeWiliProcessorType.Main, "/scripts", "WASM binary"),
        "sub": (FreeWiliProcessorType.Display, "/radio", "Radio file"),
        "fwi": (FreeWiliProcessorType.Display, "/images", "Image file"),
    }

    for ext, values in known_maps.items():
        map = FileMap.from_ext(ext)
        assert map.extension == ext
        assert map.processor == values[0]
        assert map.directory == values[1]
        assert map.description == values[2]

    with pytest.raises(ValueError, match="Extension 'failure' is not a known FreeWili file type") as _exc_info:
        FileMap.from_ext(".failure")

    assert FileMap.from_fname(r"C:\dev\My Project\Output\test.wasm") == FileMap.from_ext("wasm")
    assert FileMap.from_fname(r"/home/dev/my_project/test.wasm") == FileMap.from_ext("wasm")
    assert FileMap.from_fname(r"test.wasm") == FileMap.from_ext("wasm")

    assert FileMap.from_ext("wasm").to_path("test.wasm") == "/scripts/test.wasm"
    assert FileMap.from_ext("wasm").to_path("/some/random/path/test.wasm") == "/scripts/test.wasm"


if __name__ == "__main__":
    import pytest

    pytest.main(
        args=[
            __file__,
            "--verbose",
        ]
    )



"""Command line serial interface for the FreeWili library.

This module provides a command line interface to find and control FreeWili boards.
"""

import argparse
import importlib.metadata
import pathlib

from result import Err, Ok

from freewili import FreeWili
from freewili.cli import exit_with_error, get_device
from freewili.fw import GPIO_MAP
from freewili.serial_util import FreeWiliProcessorType, IOMenuCommand


def main() -> None:
    """A command line interface to list and control FreeWili boards.

    Parameters:
    ----------
        None

    Returns:
    -------
        None
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        default=False,
        help="List all FreeWili connected to the computer.",
    )
    parser.add_argument(
        "-i",
        "--index",
        type=int,
        default=1,
        help="Select a specific FreeWili by index. The first FreeWili is 1.",
    )
    parser.add_argument(
        "-di",
        "--display_index",
        type=int,
        default=None,
        help="Select a specific FreeWili display processor by index. The first FreeWili is 1.",
    )
    parser.add_argument(
        "-mi",
        "--main_index",
        type=int,
        default=None,
        help="Select a specific FreeWili main processor by index. The first FreeWili is 1.",
    )
    parser.add_argument(
        "-s",
        "--send_file",
        nargs=1,
        help="send a file to the FreeWili. Argument should be in the form of: <source_file>",
    )
    parser.add_argument(
        "-fn",
        "--file_name",
        nargs=1,
        help="Set the name of the file in the FreeWili. Argument should be in the form of: <file_name>",
    )
    # parser.add_argument(
    #     "-u",
    #     "--get_file",
    #     nargs=2,
    #     help="Get a file from the FreeWili. Argument should be in the form of: <source_file> <target_name>",
    # )
    parser.add_argument(
        "-w",
        "--run_script",
        nargs="?",
        const=False,
        help="Run a script on the FreeWili. If no argument is provided, -fn will be used.",
    )
    parser.add_argument(
        "-io",
        "--io",
        nargs="*",
        help=(
            "Set IO. Argument should be in the form of: <io_pin> <high/low/toggle/pwm> [pwm_freq] [pwm_duty].\n"
            "No arguments gets all IO values."
        ),
    )
    parser.add_argument(
        "-led",
        "--led",
        nargs=4,
        help="Set Board LEDs. Argument should be in the form of: <LED #> <red 0-255> <green 0-255> <blue 0-255>",
    )
    parser.add_argument(
        "-gi",
        "--gui_image",
        nargs=1,
        help="Show GUI Image. Argument should be in the form of: <image path>",
    )
    parser.add_argument(
        "-gt",
        "--gui_text",
        nargs=1,
        help="Show GUI Text. Argument should be in the form of: <text>",
    )
    parser.add_argument(
        "-rb",
        "--read_buttons",
        action="store_true",
        default=False,
        help="Read buttons.",
    )
    parser.add_argument(
        "-rd",
        "--reset_display",
        action="store_true",
        default=False,
        help="Reset the display back to the main menu.",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=importlib.metadata.version("freewili")),
    )
    args = parser.parse_args()
    device_index: int = args.index - 1
    processor_type = None
    if args.main_index is not None:
        processor_type = FreeWiliProcessorType.Main
    elif args.display_index is not None:
        processor_type = FreeWiliProcessorType.Display
    devices = FreeWili.find_all()
    if args.list:
        print(f"Found {len(devices)} FreeWili(s)")
        for i, free_wili in enumerate(devices, start=1):
            print(f"{i}. {free_wili}")
            print(f"\t{free_wili.main}")
            print(f"\t{free_wili.display}")
            print(f"\t{free_wili.ftdi}")
    if args.send_file:
        match get_device(device_index, devices):
            case Ok(device):
                file_name = None
                if args.file_name:
                    file_name = args.file_name[0]
                print(device.send_file(args.send_file[0], file_name, processor_type).unwrap())
            case Err(msg):
                exit_with_error(msg)
    # if args.get_file:
    #     match get_device(device_index, devices):
    #         case Ok(device):
    #             data = device.get_file(args.get_file[0]).unwrap()
    #             with open(args.get_file[1], "w+b") as f:
    #                 f.write(data)
    #         case Err(msg):
    #             exit_with_error(msg)
    if args.run_script is not None:
        match get_device(device_index, devices):
            case Ok(device):
                if args.run_script:
                    script_name = args.run_script
                elif args.file_name:
                    script_name = args.file_name[0]
                elif args.send_file:
                    script_name = pathlib.Path(args.send_file[0]).name
                else:
                    raise ValueError("No script or file name provided")
                print(device.run_script(script_name).unwrap())
            case Err(msg):
                exit_with_error(msg)
    if args.io is not None:
        io_args_length = len(args.io)
        if io_args_length == 0:
            match get_device(device_index, devices):
                case Ok(device):
                    print("Getting IO pin values...")
                    match device.get_io():
                        case Ok(values):
                            for io_num, io_name in GPIO_MAP.items():
                                print(f"{io_name}: {values[io_num]}")
                        case Err(msg):
                            exit_with_error(msg)
                case Err(msg):
                    exit_with_error(msg)
        else:
            io_pin: int = int(args.io[0])
            menu_cmd: IOMenuCommand = IOMenuCommand.from_string(args.io[1].lower())
            pwm_freq_hz = None
            pwm_duty_cycle = None
            if menu_cmd is IOMenuCommand.Pwm:
                io_args_length = len(args.io)
                if io_args_length < 4:
                    exit_with_error(f"Expected 4 parameters to -io, got {io_args_length}")
                pwm_freq_hz = int(args.io[2])
                pwm_duty_cycle = int(args.io[3])
            match get_device(device_index, devices):
                case Ok(device):
                    print(f"Setting IO pin {io_pin} {menu_cmd.name} ", end="")
                    if io_args_length >= 4:
                        print(f"PWM Frequency: {pwm_freq_hz}Hz {pwm_duty_cycle}%", end="")
                    print()
                    resp = device.set_io(io_pin, menu_cmd, pwm_freq_hz, pwm_duty_cycle).unwrap_or(
                        "Failed to configure IO pin"
                    )
                    if not resp.is_ok():
                        exit_with_error(f"Failed to configure IO pin: {resp.response}")
                    print(f"Successfully configured pin {io_pin} {menu_cmd.name}")
                case Err(msg):
                    exit_with_error(msg)
    if args.led:
        led_num = args.led[0]
        red = args.led[1]
        green = args.led[2]
        blue = args.led[3]
        match get_device(device_index, devices):
            case Ok(device):
                print(f"Setting LED {led_num} to RGB: {red}, {green}, {blue}...")
                match device.set_board_leds(led_num, red, green, blue):
                    case Ok(resp):
                        if not resp.is_ok():
                            exit_with_error(f"Failed to set LED {resp.response}")
                        else:
                            print(f"Successfully set LED {led_num}")
                    case Err(msg):
                        exit_with_error(msg)
            case Err(msg):
                exit_with_error(msg)
    if args.gui_image:
        value = args.gui_image[0]
        match get_device(device_index, devices):
            case Ok(device):
                print(f"Showing Image {value}...")
                match device.show_gui_image(value):
                    case Ok(resp):
                        if not resp.is_ok():
                            exit_with_error(f"Failed to show Image {resp.response}")
                        else:
                            print(f"Successfully showing {value}")
                    case Err(msg):
                        exit_with_error(msg)
            case Err(msg):
                exit_with_error(msg)
    if args.gui_text:
        value = args.gui_text[0]
        match get_device(device_index, devices):
            case Ok(device):
                print(f"Showing text {value}...")
                match device.show_text_display(value):
                    case Ok(resp):
                        if not resp.is_ok():
                            exit_with_error(f"Failed to show text {resp.response}")
                        else:
                            print(f"Successfully showing {value}")
                    case Err(msg):
                        exit_with_error(msg)
            case Err(msg):
                exit_with_error(msg)
    if args.read_buttons:
        match get_device(device_index, devices):
            case Ok(device):
                print("Getting button values...")
                match device.read_all_buttons():
                    case Ok(resp):
                        if not resp.is_ok():
                            exit_with_error(f"Failed to show text {resp.response}")
                        else:
                            print(resp.response)
                    case Err(msg):
                        exit_with_error(msg)
            case Err(msg):
                exit_with_error(msg)
    if args.reset_display:
        match get_device(device_index, devices):
            case Ok(device):
                print("Resetting display...")
                match device.reset_display():
                    case Ok(resp):
                        if not resp.is_ok():
                            exit_with_error(f"Failed to reset display {resp.response}")
                        else:
                            print(resp.response)
                    case Err(msg):
                        exit_with_error(msg)
            case Err(msg):
                exit_with_error(msg)


if __name__ == "__main__":
    main()
